/*********************************************************************
 *  Izhikevich neurons
 *  SPI to  MCP4822 dual channel 12-bit DAC output
 *  read 2 channels of ADC for input
 *  Time step 
 *********************************************************************
 * Bruce Land Cornell University
 * Oct 2016
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
/* ====== MCP4822 control word ==============
bit 15 A/B: DACA or DACB Selection bit
1 = Write to DACB
0 = Write to DACA
bit 14 ? Don?t Care
bit 13 GA: Output Gain Selection bit
1 = 1x (VOUT = VREF * D/4096)
0 = 2x (VOUT = 2 * VREF * D/4096), where internal VREF = 2.048V.
bit 12 SHDN: Output Shutdown Control bit
1 = Active mode operation. VOUT is available. ?
0 = Shutdown the selected DAC channel. Analog output is not available at the channel that was shut down.
VOUT pin is connected to 500 k???typical)?
bit 11-0 D11:D0: DAC Input Data bits. Bit x is ignored.
*/
// A-channel, 1x, active
#define DAC_config_chan_A 0b0011000000000000
#define DAC_config_chan_B 0b1011000000000000
#define two32 4294967296.0 // 2^32 

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config.h"
// threading library
#include "pt_cornell_1_2_1.h"
// for rand()
#include <stdlib.h>
volatile SpiChannel spiChn = SPI_CHANNEL2 ;	// the SPI channel to use
// for 60 MHz PB clock use divide-by-3
volatile int spiClkDiv = 2 ; // 15 MHz DAC clock

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_cmd, pt_tick, pt_Izhikevich, pt_environment ;
// uart control threads
static struct pt pt_input, pt_output, pt_DMA_output ;

// system 1 second interval tick
int sys_time_seconds ;

// ===================================================================
// == 2:14 bit fixed point 2.14 format ===============================
// ===================================================================
// resolution 2^-14 =  6.1035e-5
// dynamic range is +1.9999/-2.0
typedef signed short fix14 ;
#define multfix14(a,b) ((fix14)((((int)(a))*((int)(b)))>>14)) //multiply two fixed 2.14
#define float2fix14(a) ((fix14)((a)*16384.0)) // 2^14
#define fix2float14(a) ((float)(a)/16384.0)
// convert 14 bit to 12 bit and convert 2's comp to offset
#define fix2dac14(a)   ((int)((a)>>4)+2048)  
// #define fix2int14(a)    ((int)((a)>>14)) define ALL numbers as less than +/-1!
// #define int2fix14(a)    ((fix8)((a)<<14)) define ALL numbers as less than +/-1!
#define absfix14(a) abs(a)
#define shiftfix14(a,b) ((fix14)((int)(a)>>(int)(b)))
// generate a uniform random number between -a and +a
// NOTE: 0<a<1 range limit!
#define randfix14(a) float2fix14((-1+2*((float)rand())/RAND_MAX) * a) ;
//======================================================================
// the neurons and parameters
// n_neuron=18 is 4300 fps with learning turned ON
// n_neuron=30 is 5200 fps with learning turned OFF
#define n_neuron 10
#define v_init float2fix14(-0.65)
//================
// Izhikevich type3 neuron parameters (see bottom of program)
#define a_type3 6
#define b_type3 2
#define c_type3 float2fix14(-0.5)
#define d_type3 float2fix14(0.02)
#define Is_type3 float2fix14(0.15) //0.15
//================
// type 2 phasic spiking
#define a_type2 6
#define b_type2 2
#define c_type2 float2fix14(-0.65)
#define d_type2 float2fix14(0.06)
#define Is_type2 float2fix14(0.005) //0.15
//================
// regular spiking
#define a_type_reg 6
#define b_type_reg 2
#define c_type_reg float2fix14(-0.65)
#define d_type_reg float2fix14(0.08)
#define Is_type_reg float2fix14(0.001) //0.15
//================
// parameter arrays
fix14 v[n_neuron];
fix14 u[n_neuron];
fix14 a[n_neuron];
fix14 b[n_neuron];
fix14 c[n_neuron];
fix14 d[n_neuron];
fix14 Is[n_neuron];
// Is plus environment current 
fix14 Is_total[n_neuron];
// reset value
fix14 u_reset[n_neuron];
// boolean for spike occur
char spike[n_neuron];
// spike counter for output averaging
int cum_spike[n_neuron];
// constant 1.4 divided by 4
fix14 c14 = float2fix14(1.4/4) ;
//fix14 c54 = float2fix14(5.0/4.0) ;
// define AP peak value
fix14 peak = float2fix14(0.3) ;
// time step shift
// for dt=1/16; dt_shift=4 and dt_shift_minus2=2
// Need this because scaling required a divide by 4
#define dt_shift 4
#define dt_shift_minus2 2

//  now the synaptic weights
// total synaptic current
fix14 I_syn[n_neuron];
// first index is presyn, second is postsyn neuron 
fix14 w_syn[n_neuron][n_neuron] ;
// current positive delta weight for STDP
fix14 weight_plus[n_neuron][n_neuron] ;
// current positive delta weight for STDP
fix14 weight_minus[n_neuron][n_neuron] ;
// synaptic weight max positive weight change
fix14 d_weight_plus = float2fix14(0.0001) ;
// synaptic weight max negative weight change
// MUST be bigger than pos weight
fix14 d_weight_minus = float2fix14(0.001) ;
// max under any condition
fix14 max_weight = float2fix14(0.05) ;
// 2^4 time step time constant for STDP at 4/mSec = 4 mSec
// probably should be around 10-20 mSec
#define tau_stdp_plus 4
#define tau_stdp_minus 8
//========================
// Allow STDP learning?
// Define "learning"
//#define learning
//========================

// display variables
// DAC voltage output routing
int DAC_A_source=0, DAC_B_source=1 ;
// frame rate computation
int compute_counter, fps;

//== Timer 2 interrupt handler ===========================================
// actual scaled DAC 
volatile  int DAC_data, quadrature_data;
// ADC output
volatile  int ADC_net;	// conversion result as read from result buffer
// the DDS units:

//=============================
//void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
//{
    // 74 cycles to get to this point from timer event
   //mT2ClearIntFlag();
    
    
    // === Channel A =============
    // CS low to start transaction
     //mPORTBClearBits(BIT_0); // start transaction
    // test for ready
     //while (TxBufFullSPI2());
    // write to spi2 
    //WriteSPI2( DAC_config_chan_A | (DAC_data + 2048));
    // end SPI transaction from last interrupt cycle
    //mPORTBSetBits(BIT_0); 
 
    // main DDS phase
    //phase_accum_main += phase_incr_main  ;
    //sine_index = phase_accum_main>>24 ;
    //DAC_data = cos_table[phase_accum_main>>24]  ;
    // now the 90 degree data
    //quadrature_data = sin_table[phase_accum_main>>24];
    // read the result of channel 4 conversion from the idle buffer
    // since we are reading 10 bit integer, we can treat this as a
    // integer < 2^20 by shifting 10 bits
    //ADC_net = (ReadADC10(0)<<10) ; // max value 2^20
    //AcquireADC10(); // not needed if ADC_AUTO_SAMPLING_ON below
    
     //isr_time = ReadTimer2() ; // - isr_time;
     
//} // end ISR TIMER2

// === Serial Thread ======================================================
static PT_THREAD (protothread_cmd(struct pt *pt))
{
    // The serial interface
    static char cmd[32];
    static int iv1, iv2, iv3 ;
    static float fv1, fv2, fv3 ;
    
    PT_BEGIN(pt);
    
    // clear port used by thread 4
    // mPORTBClearBits(BIT_0);
    
      while(1) {
          
            // send the prompt via DMA to serial
            sprintf(PT_send_buffer,"cmd>");
            // by spawning a print thread
            PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
 
          //spawn a thread to handle terminal input
            // the input thread waits for input
            // -- BUT does NOT block other threads
            // string is returned in "PT_term_buffer"
            PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer(&pt_input) );
            // returns when the thead dies
            // in this case, when <enter> is pushed
            // now parse the string
            sscanf(PT_term_buffer, "%s", cmd);

            switch(cmd[0]){
                case 'w': // set weight parameters: source dest value 
                    if (cmd[1]=='w'){
                        sscanf(PT_term_buffer, "%s %d %d %f", cmd, &iv1, &iv2, &fv1); 
                        w_syn[iv1][iv2] = float2fix14(fv1);
                    }
                    if (cmd[1]=='r'){
                        sscanf(PT_term_buffer, "%s %d %d", cmd, &iv1, &iv2); 
                        sprintf(PT_send_buffer, "weight %d %d = %f\n\r", iv1, iv2, fix2float14(w_syn[iv1][iv2]));
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) ); 
                        sprintf(PT_send_buffer, "-weight %d %d = %f\n\r", iv1, iv2, fix2float14(weight_minus[iv1][iv2]));
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                        sprintf(PT_send_buffer, "+weight %d %d = %f\n\r", iv1, iv2, fix2float14(weight_plus[iv1][iv2]));
                        PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                    }
                   break;

                case 'p': // set DAC display parameters: Asource Bsource
                   sscanf(PT_term_buffer, "%s %d %d", cmd, &iv1, &iv2); 
                   DAC_A_source = iv1 ;
                   DAC_B_source = iv2 ;
                   break;
                
                case 'a': // set DAC display parameters: neuron a
                   sscanf(PT_term_buffer, "%s %d %d", cmd, &iv1, &iv2); 
                   a[iv1] = iv2 ;
                   break;
                   
                case 'b': // set DAC display parameters: neuron b
                   sscanf(PT_term_buffer, "%s %d %d", cmd, &iv1, &iv2); 
                   a[iv1] = iv2 ;
                   break;
                   
                case 'c': // set DAC display parameters: neuron c/100
                   sscanf(PT_term_buffer, "%s %d %f", cmd, &iv1, &fv2); 
                   c[iv1] = float2fix14(fv2);
                   break;
                   
                case 'd': // set DAC display parameters: neuron d/100
                   sscanf(PT_term_buffer, "%s %d %f", cmd, &iv1, &fv2); 
                   d[iv1] = float2fix14(fv2);
                   break;
                   
                case 'i': // set DAC display parameters: neuron Is/100
                   sscanf(PT_term_buffer, "%s %d %f", cmd, &iv1, &fv2); 
                   Is[iv1] = float2fix14(fv2);
                   break;
                   
                case 'r': // frame/second
                   sprintf(PT_send_buffer, "fps = %d\n\r", fps);
                   PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );
                   break;
            }
             
            // never exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread 3

// === Neural Thread ================================================

static PT_THREAD (protothread_Izhikevich(struct pt *pt))
{
    PT_BEGIN(pt);
    static int i,j ;
    // 
    #define max_rand_weight 0.000 //0.005 
    // some small, uniform, random, weights
    for(i=0; i<n_neuron; i++){  
        for(j=0; j<n_neuron; j++){
             w_syn[j][i] = randfix14(max_rand_weight) ;
             if (i==j) w_syn[j][i] = 0; // no self connection
        }
    }
    //  two bigger  weights
    //w_syn[2][1] = float2fix14(-0.07); //-0.07
    //w_syn[1][2] = float2fix14(-0.07); //-0.07
    
    // default display voltages
    DAC_A_source = 1 ;
    DAC_B_source = 0 ;
                    
    // the initial conditions
    for(i=0; i<n_neuron; i++){
        v[i] = v_init ;
        a[i] = a_type2 ;
        b[i] = b_type2 ;
        c[i] = c_type2 ;
        d[i] = d_type2 ;
        // adding  random gives a little varability between neurons
        //Is[i] = Is_type2 + randfix14(0.001) ;
        if (i==0) Is[i] = float2fix14(0.000) ;
        spike[i] = 0;
        I_syn[i] = 0;
        u_reset[i] = 0;
    }
    
    while(1) {
        //static fix14 temp ;
        compute_counter++;
        for(i=0; i<n_neuron; i++){
            // update 'voltage' of neuron
            // coded for 1/4 mSec step!
            v[i] = v[i] + (multfix14(v[i],v[i]) + v[i] + shiftfix14(v[i],2) + c14 -  
                    shiftfix14(u[i],2) + shiftfix14(Is_total[i]+I_syn[i],2)>>dt_shift_minus2);
            u_reset[i] =  u[i] + d[i];
            u[i] = u[i] + ((((v[i]>>b[i])-u[i])>>a[i])>>dt_shift);
            // 1/16 mSec step
            //v[i] = v[i] + shiftfix14(multfix14(v[i],v[i]) + v[i] + shiftfix14(v[i],2) + c14 -  
            //        shiftfix14(u[i],2) + shiftfix14(Is[i]+I_syn[i],2), 2);
            // update 'inactivation' of neuron
            //u_reset[i] =  u[i] + d[i];
            //u[i] = u[i] + ((((v[i]>>b[i])-u[i])>>a[i])>>4);
            

            // handle 'action potential peak'
            spike[i] = 0 ;
            if(v[i] > peak){
                v[i] = c[i] ;
                u[i] = u_reset[i] ;
                spike[i] = 1;
                cum_spike[i]++;
            }
        }
        
        // Compute synaptic currents
        // decay time of about 16 time steps
        for(i=0; i<n_neuron; i++){  
            // decay the current
            I_syn[i] = I_syn[i] - (I_syn[i]>>dt_shift) ;
            // 1/16
            //I_syn[i] = I_syn[i] - (I_syn[i]>>4) ;
            // add new synaptic inputs
            for(j=0; j<n_neuron; j++){
                 I_syn[i] =  I_syn[i] + ((spike[j])? w_syn[j][i] : 0) ;
                 // STDP -- can just comment out these 4 lines to eliminate learning
                 // presyn spike just before postsyn => causality => raise weight
                 // presyn spike just after  postsyn => non-causality => lower weight     
                // decay the weight increments and add in new deltas
                #ifdef learning
                 if(i!=j){
                    w_syn[j][i] = w_syn[j][i] + ((spike[i] && (w_syn[j][i] < max_weight))? weight_plus[j][i]:0) ;
                    w_syn[j][i] = w_syn[j][i] + ((spike[j] && (w_syn[j][i] > -max_weight))? weight_minus[j][i]:0) ;
                    weight_plus[j][i]  = weight_plus[j][i]  - (weight_plus[j][i]>>tau_stdp_plus)  + (spike[j]? d_weight_plus:0);
                    weight_minus[j][i] = weight_minus[j][i] - (weight_minus[j][i]>>tau_stdp_minus) - (spike[i]? d_weight_minus:0);  
                 }
                // end STDP
                #endif
            }
        }
          // some DAC output
          mPORTBClearBits(BIT_0); // start transaction
          // write to spi2 
          //WriteSPI2( DAC_config_chan_A | fix2dac14(v[DAC_A_source]) );
          WriteSPI2( DAC_config_chan_A | I_syn[0]>>4 );
          while (SPI2STATbits.SPIBUSY); // wait for end of transaction
          // end SPI transaction from last interrupt cycle
          mPORTBSetBits(BIT_0);
          //
          mPORTBClearBits(BIT_0); // start transaction
          // write to spi2 
          WriteSPI2( DAC_config_chan_B | fix2dac14(v[DAC_B_source]) );
          while (SPI2STATbits.SPIBUSY); // wait for end of transaction
          // end SPI transaction from last interrupt cycle
          mPORTBSetBits(BIT_0);
            
          PT_YIELD(pt);
          //PT_YIELD_TIME_msec(1) ;
      } // END WHILE(1)
  PT_END(pt);
} // neuron

// === the environment =============================================
// relate a measure of output to form new inputs
static PT_THREAD (protothread_environment(struct pt *pt))
{
    PT_BEGIN(pt);
      static int i;
      
      while(1) {
            // yield time 
            PT_YIELD_TIME_msec(100) ;
            // compute averate rate of  neurons
            for(i=0; i<n_neuron; i++) {
                // scale and feed back a small inhibition
               // note that avg_rate will be a small-valued integer
               // which translates to a small, fractional, positive fix14
                //I_syn[0] += cum_spike[i]*1000 ;
                Is_total[0] = Is[0] + cum_spike[i]<<5 ;
                cum_spike[i] = 0 ;
            }
            //I_syn[1] += avg_rate[n_neuron-1]*10000 ;
            //I_syn[1] -= avg_rate[n_neuron-2]*10000 ;
                    
            // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // end thread 


// === Tick ========================================================
// update a 1 second tick counter
static PT_THREAD (protothread_tick(struct pt *pt))
{
    PT_BEGIN(pt);

      while(1) {
            // yield time 1 second
            PT_YIELD_TIME_msec(1000) ;
            // compute rate of neuron simulation
            fps = compute_counter;
            compute_counter = 0 ;
            sys_time_seconds++ ;
            // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // end thread 

// === Main  ======================================================
// set up UART, timer2, threads
// then schedule them as fast as possible

int main(void)
{
  // === config the uart, DMA, vref, timer5 ISR =============
  PT_setup();

   // === setup system wide interrupts  ====================
  INTEnableSystemMultiVectoredInt();
    
  /// timer interrupt //////////////////////////
    // Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
    // 300 is 100 ksamples/sec at 30 MHz clock
    // 200 is 150 ksamples/sec
    // 150 is 200
    //OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 150);
    // set up the timer interrupt with a priority of 2
    //ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    //mT2ClearIntFlag(); // and clear the interrupt flag

    /// SPI setup //////////////////////////////////////////
    // SCK2 is pin 26 
    // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
    PPSOutput(2, RPB5, SDO2);
    // control CS for DAC
    mPORTBSetPinsDigitalOut(BIT_0);
    mPORTBSetBits(BIT_0);
    // divide Fpb by 2, configure the I/O ports. Not using SS in this example
    // 16 bit transfer CKP=1 CKE=1
    // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
    // For any given peripherial, you will need to match these
    SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV , spiClkDiv);
    
    // the ADC ///////////////////////////////////////
        // configure and enable the ADC
	CloseADC10();	// ensure the ADC is off before setting the configuration

	// define  setup parameters for OpenADC10
	// Turn module on | ouput in integer | trigger mode auto | enable autosample
        // ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
        // ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
        // ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
        #define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON 

	// define setup parameters for OpenADC10
	// ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
	#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
        //
	// Define setup parameters for OpenADC10
        // use peripherial bus clock | set sample time | set ADC clock divider
        // ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
        // ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
        // At PB clock 30 MHz, divide by two for ADC_CONV_CLK gives 66 nSec
        #define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy2

	// define setup parameters for OpenADC10
	// set AN4 and  as analog inputs
	#define PARAM4	ENABLE_AN4_ANA

	// define setup parameters for OpenADC10
	// do not assign channels to scan
	#define PARAM5	SKIP_SCAN_ALL

	// use ground as neg ref for A | use AN0 for input A
	// configure to sample AN4 
	SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN4 ); // configure to sample AN0
	OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters defined above

	EnableADC10(); // Enable the ADC
    
  // === now the threads ====================

  // init the threads
  PT_INIT(&pt_cmd);
  PT_INIT(&pt_tick);
  PT_INIT(&pt_Izhikevich);
  PT_INIT(&pt_environment);

        
  // schedule the threads
  while(1) {
    // round robin Izhikevich
    PT_SCHEDULE(protothread_cmd(&pt_cmd));
    PT_SCHEDULE(protothread_tick(&pt_tick));
    PT_SCHEDULE(protothread_Izhikevich(&pt_Izhikevich));
    PT_SCHEDULE(protothread_environment(&pt_environment));
  }
} // main

/*
 %cell parameters from Izhikevich web page
 * in the CODE: 
 * --a from this table converted to a approx right shift number so 0.02 becomes 6
 * --b from this table converted to a approx right shift number so 0.25 becomes 2 
 * --c from this table is multiplyed by 0.01 
 * --d from this table is multiplyed by 0.01
 * --Is from this table is multiplyed by 0.01
 * 
 * % a        b        c      d        Is
    0.02      0.2     -65      6       14 ;...    % tonic spiking
    0.02      0.25    -65      6       0.5 ;...   % phasic spiking
    0.02      0.2     -50      2       15 ;...    % tonic bursting -- type3
    0.02      0.25    -55     0.05     0.6 ;...   % phasic bursting
    0.02      0.2     -55     4        10 ;...    % mixed mode
    0.01      0.2     -65     8        30 ;...    % spike frequency adaptation
    0.02      -0.1    -55     6        0  ;...    % Class 1
    0.2       0.26    -65     0        0  ;...    % Class 2
    0.02      0.2     -65     6        7  ;...    % spike latency
    0.05      0.26    -60     0        0  ;...    % subthreshold oscillations
    0.1       0.26    -60     -1       0  ;...    % resonator
    0.02      -0.1    -55     6        0  ;...    % integrator
    0.03      0.25    -60     4        0;...      % rebound spike
    0.03      0.25    -52     0        0;...      % rebound burst
    0.03      0.25    -60     4        0  ;...    % threshold variability
    1         1.5     -60     0      -65  ;...    % bistability
    1       0.2     -60     -21      0  ;...    % DAP
    0.02      1       -55     4        0  ;...    % accomodation
    -0.02      -1      -60     8        80 ;...    % inhibition-induced spiking
    -0.026     -1      -45     0        80];       % inhibition-induced bursting

 */
// From http://people.ece.cornell.edu/land/courses/ece5760/DDA/NeuronIndex.htm
// Note that v and u are scaled down by a factor of 100 so that the dynamic range is -1 to +1, 
// as required by the arithmetic system and the D/A converter. 
// Compared to the equations from Iz paper, the scaled equations have all constant terms divided by 100. 
// The a and b constants are not scaled (because they are multiplied by the scaled variables), 
// but c and d are divided by 100. The constant (.04) on the v2 term above is scaled up by 100 
// because squaring (the scaled) v drops the value by 10,000. The final scaled equations are:
// v1(n+1) = v1(n) + dt*(4*(v1(n)^2) + 5*v1(n) +1.40 - u1(n) + I)
// u1(n+1) = u1 + dt*a*(b*v1(n) - u1(n)) 
// What is actually coded is:
// for a dt = 1/16 or 2>>4 (millisecond)
// from the Verilog:
	// v1(n+1) = v1(n) + dt*(4*(v1(n)^2) + 5*v1(n) +1.40 - u1(n) + I)
	// but note that what is actually coded is
	// v1(n+1) = v1(n) + (v1(n)^2) + 5/4*v1(n) +1.40/4 - u1(n)/4 + I/4)/4
	//signed_mult v1sq(v1xv1, v1, v1);
	//assign v1new = v1 + ((v1xv1 + v1+(v1>>>2) + (c14>>>2) - (u1>>>2) + (I>>>2))>>>2);
	
	// u1(n+1) = u1 + dt*a*(b*v1(n) - u1(n))
	//assign v1xb = v1>>>b;         //mult (v1xb, v1, b);
	//assign du1 = (v1xb-u1)>>>a ;  //mult (du1, (v1xb-u1), a);
	//assign u1new = u1 + (du1>>>4) ; 
	//assign u1reset = u1 + d ;